@@ -1,4 +1,5 @@
 use super::address::*;
+use super::cksumvfs;
 use super::index;
 use super::xid::*;
 use serde::{Deserialize, Serialize};
@@ -19,133 +20,137 @@ pub struct SendLogSession<'a> {
 pub struct StatCacheEntry {
     pub total_size: u64,
     pub addresses: Vec<Address>,
-    pub base_offsets: Vec<index::IndexEntryOffsets>,
+    pub data_cursors: Vec<index::RelativeDataCursor>,
     pub hashes: Vec<index::ContentCryptoHash>,
 }
 
-const SCHEMA_VERSION: i64 = 4;
+const SCHEMA_VERSION: i64 = 5;
 
 impl SendLog {
     pub fn open(p: &Path) -> Result<SendLog, anyhow::Error> {
         let mut db_conn = rusqlite::Connection::open(p)?;
+        cksumvfs::enable_sqlite_checksums(&db_conn)?;
 
-        // Only one put per send log at a time.
-        db_conn.query_row(
-            "PRAGMA locking_mode = EXCLUSIVE;",
-            rusqlite::NO_PARAMS,
-            |_r| Ok(()),
-        )?;
+        // On 64 bit platforms use sqlite3 memory mapped io.
+        if std::mem::size_of::<usize>() == 8 {
+            // 64GiB mmap size, just an estimate of the largest sendlog we are likely to see.
+            db_conn.query_row("pragma mmap_size=68719476736;", [], |_r| Ok(()))?;
+        }
+        // We want a rather large page cache for the send log.
+        // default is -2000 which is 2000 * 1024 bytes.
+        db_conn.execute("pragma cache_size = -20000;", [])?;
 
+        // Only one put per send log at a time, this simplifies checkpoints
+        db_conn.query_row("pragma locking_mode = EXCLUSIVE;", [], |_r| Ok(()))?;
         db_conn.busy_timeout(std::time::Duration::new(7 * 24 * 60 * 60, 0))?;
 
-        // Immediate lock with exclusive mode immediately blocks all other writers for the life
-        // of this open connection.
-        let tx = db_conn.transaction_with_behavior(rusqlite::TransactionBehavior::Immediate)?;
-        tx.execute(
-            "create table if not exists LogMeta(Key primary key, Value) without rowid;",
-            rusqlite::NO_PARAMS,
-        )?;
-
-        let needs_init = match tx.query_row(
+        let needs_init = match db_conn.query_row(
             "select Value from LogMeta where Key = 'schema-version';",
-            rusqlite::NO_PARAMS,
+            [],
             |r| {
                 let v: i64 = r.get(0)?;
                 Ok(v)
             },
         ) {
             Ok(v) => v != SCHEMA_VERSION,
-            Err(rusqlite::Error::QueryReturnedNoRows) => true,
-            Err(err) => return Err(err.into()),
+            Err(rusqlite::Error::SqliteFailure(err, _))
+                if err.code == rusqlite::ErrorCode::SystemIoFailure =>
+            {
+                // The failure may be due to checksumvfs being enabled
+                // for the first time, so attempt a vacuum and try again
+                // to rebuild checksums.
+                db_conn.query_row("pragma checksum_verification=OFF;", [], |_r| Ok(()))?;
+                db_conn.execute("vacuum;", [])?;
+                db_conn.query_row("pragma checksum_verification=ON;", [], |_r| Ok(()))?;
+                // Force a reinit since we don't know if our checksums were really bad.
+                true
+            }
+            Err(rusqlite::Error::SqliteFailure(err, _))
+                if err.code == rusqlite::ErrorCode::DatabaseBusy
+                    || err.code == rusqlite::ErrorCode::DatabaseLocked =>
+            {
+                anyhow::bail!("send log is busy")
+            }
+            Err(_) => true,
         };
 
+        if needs_init {
+            let tx = db_conn.transaction_with_behavior(rusqlite::TransactionBehavior::Immediate)?;
+
+            tx.execute("drop table if exists LogMeta;", [])?;
+            tx.execute("drop table if exists Sent;", [])?;
+            tx.execute("drop table if exists StatCache;", [])?;
+            tx.execute(
+                "create table LogMeta(Key primary key, Value) without rowid;",
+                [],
+            )?;
+            tx.execute(
+                "insert into LogMeta(Key, Value) values('schema-version', ?);",
+                &[&SCHEMA_VERSION],
+            )?;
+            tx.execute(
+                "insert into LogMeta(Key, Value) values('sequence-number', 1);",
+                [],
+            )?;
+            tx.execute(
+                "create table Sent(Address primary key, GCGeneration, LatestSessionId, ItemId) without rowid;",
+                [],
+            )?;
+            tx.execute(
+                "create table StatCache(Hash primary key, Cached, GCGeneration, LatestSessionId, ItemId) without rowid; ",
+                [],
+            )?;
+
+            tx.commit()?;
+
+            db_conn.execute("vacuum;", [])?;
+
+            // Final sanity check after (re)initialization.
+            let integrity_check = db_conn.query_row("pragma integrity_check;", [], |r| {
+                let v: String = r.get(0)?;
+                Ok(v)
+            })?;
+            if integrity_check != "ok" {
+                anyhow::bail!("send log integrity check failed")
+            };
+        }
+
+        let tx = db_conn.transaction_with_behavior(rusqlite::TransactionBehavior::Immediate)?;
         let sequence_number = match tx.query_row(
             "select Value from LogMeta where Key = 'sequence-number';",
-            rusqlite::NO_PARAMS,
+            [],
             |r| r.get(0),
         ) {
             Ok(n) => {
                 tx.execute(
                     "update LogMeta set Value = Value + 1 where Key = 'sequence-number';",
-                    rusqlite::NO_PARAMS,
+                    [],
                 )?;
                 n
             }
             Err(rusqlite::Error::QueryReturnedNoRows) => {
                 tx.execute(
                     "insert into LogMeta(Key, Value) values('sequence-number', 1);",
-                    rusqlite::NO_PARAMS,
+                    [],
                 )?;
                 1
             }
             Err(err) => return Err(err.into()),
         };
-
         tx.commit()?;
 
         /* Simple policy to decide when to defragment our send log. */
-        if cfg!(debug_assertions) || sequence_number % 10 == 0 {
-            db_conn.execute("vacuum;", rusqlite::NO_PARAMS)?;
+        if sequence_number % 10 == 0 {
+            db_conn.execute("vacuum;", [])?;
         }
 
-        if needs_init {
-            let mut tmp_conn = rusqlite::Connection::open(":memory:")?;
-
-            let tx = tmp_conn.transaction()?;
-
-            tx.execute(
-                "create table LogMeta(Key primary key, Value) without rowid;",
-                rusqlite::NO_PARAMS,
-            )?;
-
-            tx.execute(
-                "insert into LogMeta(Key, Value) values('schema-version', ?);",
-                &[&SCHEMA_VERSION],
-            )?;
-
-            tx.execute(
-                "insert into LogMeta(Key, Value) values('sequence-number', 1);",
-                rusqlite::NO_PARAMS,
-            )?;
-
-            tx.execute(
-                "create table Sent(Address primary key, GCGeneration, LatestSessionId, ItemId) without rowid;",
-                rusqlite::NO_PARAMS,
-            )?;
-
-            tx.execute(
-                "create table StatCache(Hash primary key, Cached, GCGeneration, LatestSessionId, ItemId) without rowid; ",
-                rusqlite::NO_PARAMS,
-            )?;
-
-            tx.commit()?;
-
-            let backup = rusqlite::backup::Backup::new(&tmp_conn, &mut db_conn)?;
-            if backup.step(-1)? != rusqlite::backup::StepResult::Done {
-                anyhow::bail!("unable to start send log transaction");
-            }
-        }
-
-        // On 64 bit platforms use sqlite3 memory mapped io.
-        if std::mem::size_of::<usize>() == 8 {
-            // 16GiB mmap size, just an estimate of the largest sendlog we are likely to see.
-            db_conn.query_row("PRAGMA mmap_size=17179869184;", rusqlite::NO_PARAMS, |_r| {
-                Ok(())
-            })?;
-        }
-
-        // We want a rather large page cache for the send log.
-        // default is -2000 which is 2000 * 1024 bytes.
-        db_conn.execute("PRAGMA cache_size = -20000;", rusqlite::NO_PARAMS)?;
-
         Ok(SendLog { db_conn })
     }
 
     pub fn session(&mut self, gc_generation: Xid) -> Result<SendLogSession, anyhow::Error> {
         // We manually control the sqlite3 transaction so we are able
         // to issue checkpoints and commit part way through a send operation.
-        self.db_conn
-            .execute("begin immediate;", rusqlite::NO_PARAMS)?;
+        self.db_conn.execute("begin immediate;", [])?;
 
         Ok(SendLogSession {
             gc_generation,
@@ -158,7 +163,7 @@ impl SendLog {
     pub fn last_send_id(&self) -> Result<Option<Xid>, anyhow::Error> {
         match self.db_conn.query_row(
             "select value from LogMeta where key = 'last-send-id';",
-            rusqlite::NO_PARAMS,
+            [],
             |r| {
                 let send_id: Xid = r.get(0)?;
                 Ok(send_id)
@@ -195,11 +200,11 @@ impl<'a> SendLogSession<'a> {
         } else {
             self.log.db_conn.execute(
                 "delete from Sent where GCGeneration != ?;",
-                &[self.gc_generation],
+                [self.gc_generation],
             )?;
             self.log.db_conn.execute(
                 "delete from StatCache where GCGeneration != ?;",
-                &[self.gc_generation],
+                [self.gc_generation],
             )?;
         }
 
@@ -211,16 +216,21 @@ impl<'a> SendLogSession<'a> {
             anyhow::bail!("no active transaction");
         };
 
-        // TODO I think we can do this in one query instead of two.
-        let has_address = self.cached_address(addr)?;
+        let mut stmt = self
+            .log
+            .db_conn
+            .prepare_cached("update Sent set LatestSessionId = ? where Address = ? returning 1;")?;
 
-        if has_address {
-            let mut stmt = self
-                .log
-                .db_conn
-                .prepare_cached("update Sent set LatestSessionId = ? where Address = ?;")?;
-            stmt.execute(rusqlite::params![self.session_id, &addr.bytes[..]])?;
-        } else {
+        let has_address = match stmt
+            .query_row(rusqlite::params![self.session_id, &addr.bytes[..]], |_r| {
+                Ok(())
+            }) {
+            Ok(_) => true,
+            Err(rusqlite::Error::QueryReturnedNoRows) => false,
+            Err(err) => return Err(err.into()),
+        };
+
+        if !has_address {
             let mut stmt = self.log.db_conn.prepare_cached(
                 "insert into Sent(GCGeneration, LatestSessionId, Address) values(?, ?, ?);",
             )?;
@@ -275,12 +285,11 @@ impl<'a> SendLogSession<'a> {
     }
 
     pub fn stat_cache_lookup(&self, hash: &[u8]) -> Result<Option<StatCacheEntry>, anyhow::Error> {
-        let mut stmt = self
-            .log
-            .db_conn
-            .prepare_cached("select Cached from StatCache where Hash = ?;")?;
+        let mut stmt = self.log.db_conn.prepare_cached(
+            "update StatCache set LatestSessionId = ?1 where Hash = ?2 returning Cached;",
+        )?;
 
-        match stmt.query_row(rusqlite::params![hash], |r| {
+        match stmt.query_row(rusqlite::params![self.session_id, hash], |r| {
             let data: Vec<u8> = r.get(0)?;
             Ok(data)
         }) {
@@ -295,11 +304,9 @@ impl<'a> SendLogSession<'a> {
             anyhow::bail!("no active transaction");
         };
 
-        self.log.db_conn.execute("commit;", rusqlite::NO_PARAMS)?;
+        self.log.db_conn.execute("commit;", [])?;
         self.tx_active = false;
-        self.log
-            .db_conn
-            .execute("begin immediate;", rusqlite::NO_PARAMS)?;
+        self.log.db_conn.execute("begin immediate;", [])?;
         self.tx_active = true;
         Ok(())
     }
@@ -336,7 +343,7 @@ impl<'a> SendLogSession<'a> {
             &[id],
         )?;
 
-        self.log.db_conn.execute("commit;", rusqlite::NO_PARAMS)?;
+        self.log.db_conn.execute("commit;", [])?;
         self.tx_active = false;
         Ok(())
     }
@@ -345,10 +352,7 @@ impl<'a> SendLogSession<'a> {
 impl<'a> Drop for SendLogSession<'a> {
     fn drop(&mut self) {
         if self.tx_active {
-            self.log
-                .db_conn
-                .execute("rollback;", rusqlite::NO_PARAMS)
-                .unwrap();
+            self.log.db_conn.execute("rollback;", []).unwrap();
         }
     }
 }
@@ -386,7 +390,7 @@ mod tests {
                     &StatCacheEntry {
                         total_size: 123,
                         addresses: vec![],
-                        base_offsets: vec![],
+                        data_cursors: vec![],
                         hashes: vec![],
                     },
                 )
@@ -454,7 +458,7 @@ mod tests {
                     &StatCacheEntry {
                         total_size: 123,
                         addresses: vec![],
-                        base_offsets: vec![],
+                        data_cursors: vec![],
                         hashes: vec![],
                     },
                 )
@@ -519,7 +523,7 @@ mod tests {
                     &StatCacheEntry {
                         total_size: 123,
                         addresses: vec![],
-                        base_offsets: vec![],
+                        data_cursors: vec![],
                         hashes: vec![],
                     },
                 )
@@ -539,7 +543,7 @@ mod tests {
                     &StatCacheEntry {
                         total_size: 123,
                         addresses: vec![],
-                        base_offsets: vec![],
+                        data_cursors: vec![],
                         hashes: vec![],
                     },
                 )
